#!/usr/bin/env python3
"""
see https://git.rtb-bl.de/ampel/acoustic/blx-doc/-/blob/master/README.md
"""

import os
import binascii
import crcmod
import struct
import io
import logging

crc_fun = None

logging.basicConfig(level=1)

# CRC16-XMODEM
crc_fun = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0x0000, xorOut=0x0000)
assert crc_fun(b'123456789') == 0x31C3, "something is wrong with the crc algo"


def rtb_serial_protocol_encode(header_preamble: str,
                               command_class: str,
                               command: str,
                               data: bytes = None) -> bytes:
    global crc_fun

    if not data:
        data = bytearray()

    res = bytearray(b'#')

    if isinstance(header_preamble, str):
        header_preamble = header_preamble[0].encode('ascii')
    res += header_preamble

    header_data_second_layer_length = 2 + 2 + len(data)

    # size of payload (2 bytes)
    res += (header_data_second_layer_length).to_bytes(2, byteorder='little')

    header_crc = crc_fun(res)
    res += header_crc.to_bytes(2, 'big')
    # print("gen header crc:", header_crc)

    if isinstance(command_class, str):
        command_class = command_class[0].encode('ascii')
    elif isinstance(command_class, int):
        command_class = bytes([command_class])
    if isinstance(command, str):
        command = command[0].encode('ascii')
    elif isinstance(command, int):
        command = bytes([command])

    res += command_class
    res += command
    res += data

    payload_crc = crc_fun(res[6:])
    res += payload_crc.to_bytes(2, 'big')
    # print("gen payload crc:", payload_crc)

    return res


def rtb_serial_protocol_parse(io_in, skip_to_start_byte=True):
    """
    Generator for parsing serial protocol packets from io streams
    """

    global crc_fun
    print("rtb_serial_protocol_parse")

    if "SKIP" in os.environ:
        io_in.seek(int(os.environ['SKIP']))

    while True:
        s = io_in.read(1)
        if len(s) == 0:
            print('end of file')
            break

        if skip_to_start_byte and s != b"#":
            # print("skipping byte (waiting for #):", s)
            continue

        preamble = io_in.read(1)

        # header data is unused

        length_bytes = io_in.read(2)
        length = int.from_bytes(length_bytes, "little") + 2
        if length > 2000:
            print("strange length:", length)
            continue

        sp_data = io_in.read(length)
        if len(sp_data) != length:
            print(f"incomplete packet, expected len: {length}, got: {len(sp_data)}")
            break

        (sp_header_crc, ) = struct.unpack(">H", sp_data[0:2])

        header_bytes = b'#' + preamble + length_bytes
        header_crc_sum = crc_fun(header_bytes)
        if header_crc_sum != sp_header_crc:
            print("RX header crc err, header bytes:", header_bytes,
                  header_crc_sum, sp_header_crc)

        sp_command_class = sp_data[2:3][0]
        sp_command = sp_data[3:4][0]
        sp_payload = sp_data[4:-2]

        crc = crc_fun(sp_data[2:])

        if crc != 0:
            print("TX Payload crc error!, crc:", hex(crc), ", data:",
                  binascii.hexlify(sp_data))
            # continue

        # print("\tsp_command_class:", hex(sp_command_class), ", ascii:", chr(sp_command_class))
        # print("\tsp_command:", hex(sp_command))
        # print("\tsp_payload len:", len(sp_payload))
        # print("\tsp_payload: 0x", binascii.hexlify(sp_payload))

        yield ((sp_command_class, sp_command, sp_payload))


def parse_serial_protocol_file(path):
    with open(path, 'rb') as f:
        yield from rtb_serial_protocol_parse(f)


if __name__ == "__main__":
    data = rtb_serial_protocol_encode('T', 'B', 'c', None)
    print("encoded:", data)

    io = io.BytesIO(data)
    for msg in rtb_serial_protocol_parse(io):
        print("msg:", msg)

    # for frame in parse_serial_protocol_file(sys.argv[1]):
    #     print("PARSE", frame)



import asyncio

class LocIdJsonRpcProtocolAsync:
    """
    Mix-in for jsonrpc abstractions
    """

    PROXIMITY_NONE = 0
    """ Numeric proximity level for 'none/'outside' range """
    PROXIMITY_NEAR = 1
    """ Numeric proximity level for 'near' range """
    PROXIMITY_FAR = 2
    """ Numeric proximity level for 'far' range """

    async def jsonrpc_request(self, cmd, params=None):
        raise BaseException("This is class in meant to be a mixin, don't use this directly. When subclassing, provide a jsonrpc_call method in super")

    async def login(self, pin):
        """
        Login with pin (required for wireless transports)
        """

        auth_state = await self.jsonrpc_request('blx.getAuthState')
        print("auth_state:", auth_state)

        if auth_state["needsPinSetup"] == True:
            return await self.jsonrpc_request('blx.login', { "pin": "" })
        else:
            return await self.jsonrpc_request('blx.login', { "pin": pin })

    async def ping(self, echo=None):
        """
        Ping locid module
        """
        return await self.jsonrpc_request('locid.ping', { "echo": echo })

    async def restart(self, ):
        """
        Restarts firmware (3 seconds after call).
        """
        return await self.jsonrpc_request('locid.restart')

    async def getVersion(self, ):
        """
        Gets firmware version information
        """
        return await self.jsonrpc_request('locid.getVersion')

    async def getStatus(self, ):
        """
        Gets status (connected users, activation state, ...)
        """
        return await self.jsonrpc_request('locid.getStatus')

    async def getBeaconId(self, ):
        """
        Gets the unique beacon id of this module
        """
        return await self.jsonrpc_request('locid.getBeaconId')

    async def getMetaInfo(self, ):
        """
        Reads meta information descriptor (raw)
        """
        return await self.jsonrpc_request('locid.getMetaInfo')

    async def setMetaInfo(self, metaInfo=None):
        """
        Writes meta information descriptor (raw)
        """
        return await self.jsonrpc_request('locid.setMetaInfo', { "metaInfo": metaInfo })

    async def getNear(self, ):
        """
        Gets near range threshold
        """
        return await self.jsonrpc_request('locid.getNear')

    async def setNear(self, near=None):
        """
        Sets near range threshold
        """
        return await self.jsonrpc_request('locid.setNear', { "near": near })

    async def getFar(self, ):
        """
        Gets far range threshold
        """
        return await self.jsonrpc_request('locid.getFar')

    async def setFar(self, far=None):
        """
        Sets far range threshold
        """
        return await self.jsonrpc_request('locid.setFar', { "far": far })

    async def getBeaconRSSI(self, ):
        """
        Gets beacon calibration value
        """
        return await self.jsonrpc_request('locid.getBeaconRSSI')

    async def setBeaconRSSI(self, rssi:int=None):
        """
        Sets beacon calibration value (RSSI in 1m distance)
        """
        return await self.jsonrpc_request('locid.setBeaconRSSI', { "rssi": rssi })

    async def getTxPower(self, ):
        """
        Gets bluetooth transmission power
        """
        return await self.jsonrpc_request('locid.getTxPower')

    async def setTxPower(self, txpower:int=None):
        """
        Sets bluetooth transmission power
        """
        allowed = [-40, -20, -16, -12, -8, -4, 0, 3, 4, 8]
        if not txpower in allowed:
            raise BaseException(f"Invalid arg, allowed: {allowed}")

        return await self.jsonrpc_request('locid.setTxPower', { "txpower": txpower })

    async def enable_advanced_stats(self, enable):
        """
        Enables advanced debugging statistics (for locid.getStatus)
        """

        return await self.jsonrpc_request("locid.setEnableAdvancedStatistics", {"enable": enable})


import asyncio
import logging
from typing import Dict, Union, List
from numbers import Number
import json
import base64

logger = logging.getLogger('json_rpc_io')
logger.setLevel(logging.INFO)

RFC7464_START = 0x1E
RFC7464_END = 0x0A

class JsonRpcError(BaseException):
    pass

JsonRpcParam = Union[str, int, float, Dict[str, any], List[any]]
JsonRpcResult = Union[str, int, float, Dict[str, any], List[any]]

class JsonRpcIO:
    """ Generic implemention of a json rpc protocol over an (asyncio based) serial transport
    (Framing is currently forced to be in RFC7464 format!).
    """

    rx_task: asyncio.Task
    requests_futures: Dict[any, asyncio.Future]
    req_id: int

    def __init__(self, io):
        self.io = io
        self.req_id = 0
        self.requests_futures = {}

        self.rx_task = asyncio.create_task(self._rx_task_exec())

    async def __aenter__(self):
        logging.debug("jsonrpcio aenter")
        await self.io.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        logging.debug("jsonrpcio aexit")
        await self.io.close()
        self.rx_task.cancel()

        try:
            await self.rx_task
        except asyncio.CancelledError:
            logging.debug("rx_task is cancelled now")

    async def _rx_task_exec(self):
        while not self.rx_task.done():
            line = await self.io.readline()
            logging.debug("JSONRPC GOT LINE: {line}")

            if line[0] != RFC7464_START or line[-1] != RFC7464_END:
                raise "malformed rx (missing framing)"

            json_data = line[1:-1]
            j = json.loads(json_data)

            await self._handle_response(j)

    async def _handle_response(self, j: Dict[str, any]):
        if "error" in j:
            id = j["id"]
            future = self.requests_futures[id]
            if future:
                future.set_exception(JsonRpcError(f"err: {j}"))
                del self.requests_futures[id]
        elif "result" in j:
            id = j["id"]
            future = self.requests_futures[id]
            if future:
                future.set_result(j['result'])
                del self.requests_futures[id]
        else:
            raise("todo: jsonrpc notification:", j)

    async def jsonrpc_request(self, method: str, params: JsonRpcParam=None, timeout=10, auto_encode_b64=True) -> JsonRpcResult:
        """ Sends a json rpc request and waits for the result """

        cmd = {
             "method": method,
            "id": self.req_id,
        }

        if params:
            if auto_encode_b64:
                for key, val in params.items():
                    if isinstance(val, bytes) or isinstance(val, bytearray):
                        params[key] = base64.b64encode(val).decode('ascii')

            cmd["params"] = params

        self.req_id += 1

        data = bytes([RFC7464_START]) + json.dumps(cmd).encode() + bytes([RFC7464_END])
        await self.io.write(data)

        fut = asyncio.get_running_loop().create_future()
        self.requests_futures[cmd["id"]] = fut

        try:
            return await asyncio.wait_for(fut, timeout=timeout)
        except asyncio.TimeoutError as ex:
            logging.exception("future wait_for")
            raise JsonRpcError(f"Timeout for request: {cmd} after {timeout}s") from ex

    # compat (remove someday)
    async def request(self, method, params=None, auto_encode_b64=True):
        return await self.jsonrpc_request(method, params, auto_encode_b64)



#!/usr/bin/env python3

import sys
import inspect
import asyncio
import argparse
import time
import json

from .ble import *
from .ble import LocIdBle
from .locid import LocIdUSBDongle


async def main_async():

    parser = argparse.ArgumentParser(
                    prog='locid',
                    description='locid command line utility')

    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.add_parser('hci-ble-scan', help="Bluetooth scan & dump nearby locid devices (using HCI adapter)")
    subparsers.add_parser('watch-state', help="Watches state (polling)")
    parser_usb = subparsers.add_parser('raw-exec', help="Runs methods on USB-Dongle from command line (e.g. 'restart')")
    subparsers.add_parser('raw-event-dump', help="Dumps raw events")
    subparsers.add_parser('mesh-state', help="Dumps state of mesh/net2")
    subparsers.add_parser('mesh-scan', help="Scans for mesh/net2 devices in network")

    parser_usb.add_argument("method")
    parser_usb.add_argument("params", nargs="*")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    # print("argparse res:", args)

    if args.cmd == "ble-scan":
        await LocIdBle.scan_print(timeout=30)
        return

    elif args.cmd == 'raw-exec':
        with LocIdUSBDongle() as dev:
            v = dev.getVersion()["version"]
            print(f"USB Dongle fw version: {v}")
            print(f"Beacon ID: {dev.getBeaconId().get('id'):#0{8}x}")

            # print("available dongle methods:", dir(l))

            method_name = args.method
            try:
                method = getattr(dev, method_name)
            except AttributeError:
                print("no such method: ", method_name)
                sys.exit(1)

            sig = inspect.signature(method)
            print("signature:", sig)

            params = args.params

            params_t = []
            for i, p in enumerate(params):
                sigp = list(sig.parameters.values())[i]
                type = sigp.annotation
                if type == int:
                    params_t.append(int(p))
                else:
                    params_t.append(p)

            print("params after map:", params_t)

            res = method(*params_t)
            print("result:", res)

    elif args.cmd == "raw-event-dump":
        with LocIdUSBDongle() as dev:

            def on_notification(n):
                print("Raw notification:", n)
            dev.on_notification(on_notification)

            while True:
                time.sleep(1)

    elif args.cmd == "watch-state":
        with LocIdUSBDongle() as dev:

            def on_notification(n):
                print("notification: ", n)

            dev.on_notification(on_notification)

            v = dev.getVersion()["version"]
            print(f"USB Dongle firmware version: {v}")

            while True:
                s = dev.getStatus()
                print(s)
                time.sleep(1)

    elif args.cmd == "mesh-state":
        with LocIdUSBDongle() as dev:
            r = dev.jsonrpc_call("mesh.getStatus")
            print(json.dumps(r, indent=4))

    elif args.cmd == "mesh-scan":
        with LocIdUSBDongle() as dev:

            def on_notification(n):
                # print("notification: ", n)
                pass

            dev.on_notification(on_notification)

            print("Requesting meta information...")
            dev.jsonrpc_call("mesh.requestMeta")

            time.sleep(5)

            nodes_res = dev.jsonrpc_call("mesh.getNodes")

            for n in nodes_res['nodes']:
                addr = n['addr']
                node_stats = dev.jsonrpc_call("mesh.getNodeStats", {"addr": addr})
                print("*" * 100)
                print("mesh node:", json.dumps(node_stats, indent=4))


def main():
    """entry point (when installed by pip as script)"""
    asyncio.run(main_async())



#!/usr/bin/env python3

import json
import time
import os
import logging
from threading import Thread, Condition, Event

import serial
import serial.tools.list_ports

from .locid_jsonrpc_proto_sync import LocIdJsonRpcProtocol

logger = logging.getLogger('locid_usb')
logger.setLevel(logging.DEBUG)

LOCID_USB_VID = 21018
LOCID_USB_PID = 1


class LocIdError(BaseException):
    pass


class JsonRPCError(LocIdError):
    pass


class TransportError(LocIdError):
    pass


def find_serial_port() -> str:
    for port in serial.tools.list_ports.comports():
        logging.debug(f"port: {port}, path: {port.device}, vid: {port.vid}, pid: {port.pid}")

        if port.vid == LOCID_USB_VID and port.pid == LOCID_USB_PID:
            return port.device

    return None


class LocIdSerialTransportThread(Thread):

    def __init__(self, device_port: str = None):
        """ Initializes and connects LocId serial transport instance.

        :param device_port: Serial port of locid module. Use e.g. "COMX" on windows, "/dev/ttyUSB*" on unix. If omitted, the device is tried to be determined automatically.
        """

        Thread.__init__(self)
        self.daemon = True
        self.device_port = device_port
        self.serial = None

        self._jsonrpc_id = 0
        self._error_callback = None
        self._status_callback = None
        self._notification_callback = None
        self._usb_forward_callback = None
        self._user_connected_callback = None
        self._user_disconnected_callback = None

        self.jsonrpc_conds = dict()
        self.jsonrpc_responses = dict()

        self._users_by_id = {}
        self._stop_rx_thread_evt = Event()

    def connect_serial(self):
        if not self.device_port:
            self.device_port = os.environ.get('DEV')
        if not self.device_port:
            self.device_port = find_serial_port()
        if not self.device_port:
            raise TransportError("Couldn't auto-detect serial interface (use DEV environment variable, e.g. DEV=COM1 or DEV=/dev/ttyUSB0)")

        logging.info("using device port: %s", self.device_port)
        self.serial = serial.Serial(self.device_port)
        self.start()

    def connect(self):
        self.connect_serial()

    def close(self):
        logging.debug("close")
        self._stop_rx_thread_evt.set()
        self.serial.close()
        self.join()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def jsonrpc_call(self, method, params=None, ignore_response=False):
        """ Sends a locid protocol request.
        (see https://locid.org/dev/jsonrpc-doc/)

        :param method: method name (e.g. locid.ping)
        :param params: parameter dictionary (optional)
        :param ignore_response: If True, this call won't block and ignore results
        :returns: result object
        """

        if self.serial is None:
            logging.warning("please call connect() explicitly before")
            self.connect_serial()

        # see https://www.jsonrpc.org/specification
        jsonrpc_id = self._jsonrpc_id
        self._jsonrpc_id += 1

        request_j = {
            "id": jsonrpc_id,
            "method": method,
        }
        if params:
            request_j['params'] = params

        lock = Condition()
        self.jsonrpc_conds[jsonrpc_id] = lock

        bytes = (json.dumps(request_j) + "\n").encode()
        logging.debug("sending bytes: %s, len: %d", bytes, len(bytes))
        self.serial.write(bytes)
        self.serial.flush()

        if ignore_response:
            return None

        with lock:
            lock.wait(10)

        if jsonrpc_id not in self.jsonrpc_responses:
            raise JsonRPCError(f"no response for jsonrpc request {jsonrpc_id}")

        res_j = self.jsonrpc_responses[jsonrpc_id]

        del self.jsonrpc_responses[jsonrpc_id]

        if "error" in res_j:
            raise JsonRPCError(f"jsonrpc error: {res_j} (request: {request_j})")
        else:
            return res_j.get("result")

    def reset(self):
        """ Restarts firmware """
        self._stop_rx_thread_evt.set()
        res = self.jsonrpc_call("locid.restart", ignore_response=True)
        return res

    def dfu(self):
        """ Enters DFU mode """

        logging.info("stopping tx thread...")
        self._stop_rx_thread_evt.set()
        res = self.jsonrpc_call("locid.enterUsbDfu", ignore_response=True)
        time.sleep(3)
        logging.info("closing serial")
        self.serial.close()
        return res

    def on_status(self, callback):
        self._status_callback = callback

    def on_error(self, callback):
        self._error_callback = callback

    def on_client_request(self, callback):
        self._usb_forward_callback = callback

    def on_notification(self, callback):
        self._notification_callback = callback

    def on_user_min_prox_changed(self, old, prox):
        logging.info("on_user_min_prox_changed", old, prox)

    def on_user_connected(self, user_id):
        logging.info("on_user_connected: %s", user_id)
        if self._user_connected_callback:
            self._user_connected_callback(self, user_id)
        pass

    def on_user_disconnected(self, user_id):
        logging.info("on_user_disconnected: %s", user_id)
        if self._user_disconnected_callback:
            self._user_disconnected_callback(self, user_id)
        pass

    def on_user_proximity_changed(self, user_id):
        pass

    def handle_status_notification(self, status):
        # print("handle_status_notification:", status)

        if "users" in status:
            connected_before = set(filter(lambda u: self._users_by_id[u].get('connected', False), self._users_by_id.keys()))
            # print("connected_before:", connected_before)

            for u in self._users_by_id.values():
                u['connected'] = False

            for u in status["users"]:
                uid = u["id"]
                if uid not in self._users_by_id:
                    self._users_by_id[uid] = u

                self._users_by_id[uid]['connected'] = True

                if uid not in connected_before:
                    self.on_user_connected(uid)

            for uid in connected_before:
                if not self._users_by_id[uid]['connected']:
                    self.on_user_disconnected(uid)

        if self._status_callback:
            self._status_callback(status)

    def run(self):
        while not self._stop_rx_thread_evt.is_set():
            # print("waiting for line...")
            line = None
            try:
                line = self.serial.readline()
            except serial.serialutil.SerialException as e:
                stopped = self._stop_rx_thread_evt.is_set()
                if stopped:
                    break
                else:
                    logger.exception(f"Serial ex, stop: {stopped}")
                    raise TransportError("serial disconnect") from e

            logging.debug("<<< %s", line)

            j = None
            try:
                j = json.loads(line)
            except json.decoder.JSONDecodeError as e:
                logging.error("json dec error: %s", e)
                print(line)
                if self._error_callback:
                    self._error_callback(e)
                continue

            # logging.debug("json: %s", j)

            # jsonrpc responses (and errors)
            if 'id' in j:
                self.jsonrpc_responses[j['id']] = j
                cond = self.jsonrpc_conds[j['id']]
                with cond:
                    cond.notify()
                del self.jsonrpc_conds[j['id']]

            # jsonrpc notifications
            elif 'method' in j:
                method = j['method']
                if method == 'locid.status':
                    self.handle_status_notification(j['result'])

                # elif method == 'locid.user.prox_changed':
                #     print("TODO")

                elif method == 'locid.usbForward' or method == 'locid.hostForward':
                    logging.debug("locid.usbForward: %s", j)

                    data = j.get("params", {}).get("data")
                    if data is None:
                        logging.error("missing data arg")
                    else:
                        if self._usb_forward_callback:
                            self._usb_forward_callback(data)

                else:
                    if self._notification_callback:
                        self._notification_callback(j)
                    else:
                        logger.warning("Unhandled notification (set notification_callback property to handle): %s", j)

        logging.debug("thread exit")


class LocIdUSBDongle(LocIdSerialTransportThread, LocIdJsonRpcProtocol):

    @classmethod
    def all(cls):
        res = []
        for port in serial.tools.list_ports.comports():
            logging.debug(f"port: {port}, path: {port.device}, vid: {port.vid}, pid: {port.pid}")

            if port.vid == LOCID_USB_VID and port.pid == LOCID_USB_PID:
                res.append(LocIdUSBDongle(port.device))

        return res


# compat
class LocId(LocIdUSBDongle):
    """ Deprecated: Use LocIdUSBDongle """
    pass




class LocIdJsonRpcProtocol:
    """
    Kept for compatibility. Use LocIdJsonRpcProtocolAsync!
    """

    PROXIMITY_NONE = 0
    """ Numeric proximity level for 'none/'outside' range """
    PROXIMITY_NEAR = 1
    """ Numeric proximity level for 'near' range """
    PROXIMITY_FAR = 2
    """ Numeric proximity level for 'far' range """

    def jsonrpc_call(self, cmd, params=None):
        raise BaseException("This is class in meant to be a mixin, don't use directly")

    def ping(self, echo=None):
        """
        Ping locid module
        """
        return self.jsonrpc_call('locid.ping', { "echo": echo })

    def restart(self):
        """
        Restarts firmware (3 seconds after call).
        """
        return self.jsonrpc_call('locid.restart')

    def getVersion(self):
        """
        Gets firmware version information
        """
        return self.jsonrpc_call('locid.getVersion')

    def getStatus(self):
        """
        Gets status (connected users, activation state, ...)
        """
        return self.jsonrpc_call('locid.getStatus')

    def getState(self):
        """
        Gets state (connected users, activation state, ...)
        """
        return self.jsonrpc_call('locid.getState')

    def getBeaconId(self):
        """
        Gets the unique beacon id of this module
        """
        return self.jsonrpc_call('locid.getBeaconId')

    def getMetaInfo(self):
        """
        Reads meta information descriptor (raw)
        """
        return self.jsonrpc_call('locid.getMetaInfo')

    def setMetaInfo(self, metaInfo=None):
        """
        Writes meta information descriptor (raw)
        """
        return self.jsonrpc_call('locid.setMetaInfo', { "metaInfo": metaInfo })

    def getNear(self):
        """
        Gets near range threshold
        """
        return self.jsonrpc_call('locid.getNear')

    def setNear(self, near=None):
        """
        Sets near range threshold
        """
        return self.jsonrpc_call('locid.setNear', { "near": near })

    def getFar(self):
        """
        Gets far range threshold
        """
        return self.jsonrpc_call('locid.getFar')

    def setFar(self, far=None):
        """
        Sets far range threshold
        """
        return self.jsonrpc_call('locid.setFar', { "far": far })

    def getBeaconRSSI(self):
        """
        Gets beacon calibration value
        """
        return self.jsonrpc_call('locid.getBeaconRSSI')

    def setBeaconRSSI(self, rssi:int=None):
        """
        Sets beacon calibration value (RSSI in 1m distance)
        """
        return self.jsonrpc_call('locid.setBeaconRSSI', { "rssi": rssi })

    def getTxPower(self):
        """
        Gets bluetooth transmission power
        """
        return self.jsonrpc_call('locid.getTxPower')

    def setTxPower(self, txpower:int=None):
        """
        Sets bluetooth transmission power
        """
        allowed = [-40, -20, -16, -12, -8, -4, 0, 3, 4, 8]
        if not txpower in allowed:
            raise BaseException(f"Invalid arg, allowed: {allowed}")

        return self.jsonrpc_call('locid.setTxPower', { "txpower": txpower })

    def enable_advanced_stats(self, enable):
        """
        Enables advanced debugging statistics (for locid.getStatus)
        """

        return self.jsonrpc_call("locid.setEnableAdvancedStatistics", {"enable": enable})


"""
Stuff for interacting with locid devices via bluetooth
"""

import asyncio
import json
from typing import Union, Tuple, Dict, List, AsyncGenerator
import logging
import bleak

from bleak import BleakClient, BleakScanner, BleakGATTCharacteristic
from pprint import pprint

logger = logging.getLogger('locid_ble')
# logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)

LOCID_BLE_SERVICE_UUID = 'b940d010-f5f8-466e-aff9-25556b57fe6d'

LOCID_BLE_UUID_CHAR_FIRMWARE_VERSION =    0x0011
LOCID_BLE_UUID_CHAR_DISTANCE =       0xD011
LOCID_BLE_UUID_CHAR_NEAR =           0xD012
LOCID_BLE_UUID_CHAR_FAR =            0xD013
LOCID_BLE_UUID_CHAR_MAC =            0xD014
LOCID_BLE_UUID_CHAR_TIME =           0xD015
LOCID_BLE_UUID_CHAR_ENABLE=          0xD016
LOCID_BLE_UUID_CHAR_LONG =           0xD017
LOCID_BLE_UUID_CHAR_LAT =            0xD018
LOCID_BLE_UUID_CHAR_DOOR =           0xD019
LOCID_BLE_UUID_CHAR_INVERT_DOOR =    0xD020
LOCID_BLE_UUID_CHAR_GROUP_ID =       0xD021
LOCID_BLE_UUID_CHAR_SYSTEM_ID =       0xD022
LOCID_BLE_UUID_CHAR_BEACON_RSSI =     0xD023
LOCID_BLE_UUID_CHAR_APS_CALL =		0xD024
LOCID_BLE_UUID_CHAR_HOST_COMM =		0xD025
LOCID_BLE_UUID_CHAR_DEVICE_META_INFO =		0xD026
LOCID_BLE_UUID_CHAR_NAME =           0xD01F

def locid_char_uuid_128(short):
    hexstr = "%0.4x" % short
    return f"b940{hexstr}-f5f8-466e-aff9-25556b57fe6d"

async def _bleak_client_conn(client: BleakClient, n_retries=5, disconnected_callback=None):
    # connect
    last_err = None
    for attempt in range(n_retries):
        try:
            await client.connect(
                dangerous_use_bleak_cache=True,
                timeout=20,
                disconnected_callback=disconnected_callback
                )
            return client
        except (bleak.exc.BleakError, bleak.exc.BleakDeviceNotFoundError, bleak.exc.BleakDBusError, asyncio.exceptions.TimeoutError) as e:
            logger.error(f"bleak con err: '{e}' ({type(e)})")
            logger.error(f"discovered devs: {BleakScanner.discovered_devices}")
            if attempt < n_retries:
                print("retrying...")
            last_err = e
            await asyncio.sleep(2)
            #client = BleakClient(client.address)

    raise last_err


async def bleak_client_reliable_connect(client_or_addr: Union[BleakClient, str], n_retries=5, disconnected_callback=None) -> BleakClient:
    """
    Helper for getting connected BleakClient instances (using multiple attempts / platform workarounds)
    """

    if not disconnected_callback:
        disconnected_callback = lambda d: print("discon:", d)

    client = None
    if isinstance(client_or_addr, BleakClient):
        client = client_or_addr
        await _bleak_client_conn(client, n_retries, disconnected_callback)

    elif isinstance(client_or_addr, str):
        logging.info(f"Searching for locid device with ble mac: {client_or_addr}")

        for _ in range(3):
            try:
                bledev = await BleakScanner.find_device_by_address(client_or_addr, timeout=5, cb={"use_bdaddr": True})
                print("bledev:", bledev)
                if bledev:
                    client = BleakClient(bledev)

                await _bleak_client_conn(client, n_retries, disconnected_callback)
                return
            except bleak.exc.BleakError as e:
                logging.exception("bleak error (scan/conn)")
                await asyncio.sleep(2)

    else:
        raise BaseException("invalid client arg")

    if client is None:
        raise BaseException(f"Couldn't find client with addr: {client_or_addr}")


BLE_SERVICE_UUID_NORDIC_NUS = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"
BLE_CHAR_UUID_NORDIC_NUS_TX = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
BLE_CHAR_UUID_NORDIC_NUS_RX = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"

class BleakNordicNusError(BaseException):
    pass

class BleakNordicNus(asyncio.StreamReader, asyncio.StreamWriter):

    client: BleakClient
    client_addr: str
    rxbuf: bytearray

    MAX_RXBUF_LEN = 4096

    def __init__(self, client: Union[BleakClient, str]):

        if client is BleakClient:
            self.client = client
            self.client_addr = client.address
        else:
            self.client_addr = client
            self.client = BleakClient(client)

        self.rxbuf = bytearray()
        self.read_future = None

        super().__init__()

    def on_disconnect(self, bleak_dev: BleakClient):
        logger.info(f"DISCONNECT, bleak_dev: {bleak_dev}")

    async def connect(self):
        # if self.client is None:
        #     self.client = await bleak_client_reliable_connect(self.client_addr, disconnected_callback=self.on_disconnect)

        if not self.client.is_connected:
            logging.info(f"connecting bleak client: {self.client}")
            await bleak_client_reliable_connect(self.client)

        try:
            await self.client.start_notify(BLE_CHAR_UUID_NORDIC_NUS_RX, self._on_rx)
        except Exception as ex:
            logger.exception("notifiy rx char")
            logger.error(f"is connected: {self.client.is_connected}")
            await self.client.disconnect()
            raise ex

    @classmethod
    async def open_connection(cls, client_or_addr) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        self = cls(client_or_addr)
        await self.connect()
        return (self, self)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()

    async def disconnect(self):
        if self.read_future:
            self.read_future.cancel()
            self.read_future = None

        if self.client.is_connected:
            try:
                await self.client.stop_notify(BLE_CHAR_UUID_NORDIC_NUS_RX)
            except bleak.exc.BleakError:
                logger.warning("stop notify error (probably bleak bug):", exc_info=1)

            await self.client.disconnect()
        else:
            logger.warning("already disconnected")

    async def close(self):
        await self.disconnect()

    async def write(self, data, mtu=200):
        logger.debug(f"write, data: {data}")

        if not self.client.is_connected:
            raise BleakNordicNusError("client not connected. Please call connect() first")

        data_len = len(data)
        chunks = [ data[i:i + mtu] for i in range(0, data_len, mtu) ]

        for chunk in chunks:
            await self.client.write_gatt_char(BLE_CHAR_UUID_NORDIC_NUS_TX, chunk, response=True)

    async def readline(self):
        logger.debug("readline()")

        fut = asyncio.get_running_loop().create_future()
        self.read_future = fut
        await self._refresh_rx()
        return await fut

    async def _refresh_rx(self):
        """Parses rx buffer and calls callbacks/futures for complete commands"""

        if self.read_future:
            end_idx = self.rxbuf.find(b"\n")

            if end_idx >= 0:
                line = self.rxbuf[0:end_idx+1]
                self.rxbuf = self.rxbuf[end_idx+1:]

                self.read_future.set_result(line)
                self.read_future = None
        else:
            logging.error("no read future on _refresh_rx")

    async def _on_rx(self, sender: BleakGATTCharacteristic, data: bytearray):
        logger.debug(f"on_rx, data: {data} (sender: {sender}, )")

        self.rxbuf += data
        logger.debug(f"rxbuf: {self.rxbuf}")

        if len(self.rxbuf) > self.__class__.MAX_RXBUF_LEN:
            raise BaseException("rxbuf overflow")

        await self._refresh_rx()

from .json_rpc_io import JsonRpcIO
from .locid_jsonrpc_proto_async import LocIdJsonRpcProtocolAsync

class BleJsonRpcClient(JsonRpcIO):
    def __init__(self, bleak_client_or_addr: Union[BleakClient, str]):
        assert bleak_client_or_addr is not None

        nus = BleakNordicNus(bleak_client_or_addr)
        super().__init__(nus)

    async def __aenter__(self):
        print("BleJsonRpcClient AENTER")
        return await super().__aenter__()

class LocIdBleClient(BleJsonRpcClient, LocIdJsonRpcProtocolAsync):

    @property
    def bleak_client(self) -> BleakClient:
        """Accessor for the underlying bleak client"""

        nus: BleakNordicNus = self.io
        return nus.client

    async def connect(self):
        await self.io.connect()

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.bleak_client.address}>"

    def __repr__(self) -> str:
        return self.__str__()

class LocIdBle:

    @classmethod
    async def scan_devices(cls, timeout=5) -> AsyncGenerator[LocIdBleClient, None]:
        """Scans for supported locid bluetooth devices

        (This is an async generator, use with 'async for ...' statement)

        Args:
            timeout: Time to scan for (in seconds)
        """

        queue = asyncio.Queue()

        # duplicate filter
        reported_macs = set()

        def on_det(bleak_dev: BleakClient, advert):
            if not bleak_dev.address in reported_macs:
                dev = LocIdBleClient(bleak_dev)
                queue.put_nowait(dev)
                reported_macs.add(bleak_dev.address)

        bleak_dis_fut = BleakScanner.discover(
            timeout=timeout,
            detection_callback=on_det,
            return_adv=True,
            cb={ "use_bdaddr": True },
            service_uuids=[ LOCID_BLE_SERVICE_UUID ],
            )

        bleak_disc_task = asyncio.create_task(bleak_dis_fut)

        while not bleak_disc_task.done():
            qget = queue.get()
            try:
                dev = await asyncio.wait_for(qget, timeout=1)
                yield dev
            except asyncio.TimeoutError:
                continue

        await asyncio.wait_for(bleak_disc_task, timeout=timeout+1)

    @classmethod
    async def devices(cls, timeout=3) -> List['LocIdBleClient']:
        """Scans for supported locid bluetooth devices and returns them as a list

        Args:
            timeout: Time to scan for (in seconds)
        """

        return [i async for i in LocIdBle.scan_devices(timeout=timeout)]

    @classmethod
    async def scan_print(cls, services=False, timeout=10):

        async for dev in LocIdBle.scan_devices(timeout=timeout):
            print("Found device:", dev)

            try:
                await dev.connect()

                bleak_client = dev.bleak_client

                print("*" * 100)
                print(f"MAC: {bleak_client.address}")

                name_raw = await bleak_client.read_gatt_char(locid_char_uuid_128(LOCID_BLE_UUID_CHAR_NAME))
                name = name_raw.decode()
                print(f"Name (char): {name}")

                fw_raw = await bleak_client.read_gatt_char(locid_char_uuid_128(LOCID_BLE_UUID_CHAR_FIRMWARE_VERSION))
                fw = fw_raw.decode()
                print(f"Firmware: {fw}")

                if services:
                    for service in bleak_client.services:
                        print("\tservice:", service)
                        for char in service.characteristics:
                            print("\t\tchar:", char)

            except Exception as ex:
                print("ex:", ex)
